# $Header: /usr/custom/cvs/public/EasyIndexes/EasyKeywordIndex.py,v 1.1 2002/07/06 14:37:01 johanc Exp $
#
# (c) 2001, Torped Strategi och Kommunikation AB.
# See License.txt for License information.
#

__doc__="""EasyKeywordIndex"""
__version__='$Revision: 1.1 $'[11:-2]


from zLOG import LOG, ERROR
from types import StringType
from BTrees.OOBTree import OOSet, difference


from Globals import Persistent, DTMLFile
from Acquisition import Implicit
from OFS.History import Historical
from OFS.SimpleItem import SimpleItem

from Products.PluginIndexes import PluggableIndex 
from Products.PluginIndexes.common.UnIndex import UnIndex 
from Products.PluginIndexes.common.util import parseIndexRequest
from BTrees.IIBTree import IITreeSet, IISet #, union, intersection, difference
from BTrees.IIBTree import union as IIunion
from BTrees.IIBTree import intersection as IIintersection 
from BTrees.IIBTree import difference as IIdifference 

from types import StringType, ListType, IntType, TupleType
_marker = []

class EasyKeywordIndex(UnIndex,PluggableIndex.PluggableIndex,Persistent,Implicit,SimpleItem):

    __implements__ = (PluggableIndex.PluggableIndexInterface,)

    meta_type="EasyKeywordIndex"

    manage_options= (
        {'label': 'Settings',     
         'action': 'manage_main',
        },
    )

    query_options = ["query","operator"]

    #NOTPATCH
    def __init__(self, id, ignore_ex=None, call_methods=None):
        UnIndex.__init__(self, id, ignore_ex=ignore_ex, call_methods=call_methods)
        self.operators = ['or','and','not','notor','notand']  #NOTPATCH
        self.useOperator = 'or'
    #NOTPATCH    
        
    """Like an UnIndex only it indexes sequences of items
    
    Searches match any keyword.

    This should have an _apply_index that returns a relevance score
    """

    def index_object(self, documentId, obj, threshold=None):
        """ index an object 'obj' with integer id 'i'

        Ideally, we've been passed a sequence of some sort that we
        can iterate over. If however, we haven't, we should do something
        useful with the results. In the case of a string, this means
        indexing the entire string as a keyword."""

        # First we need to see if there's anything interesting to look at
        # self.id is the name of the index, which is also the name of the
        # attribute we're interested in.  If the attribute is callable,
        # we'll do so.

        newKeywords = getattr(obj, self.id, None)
        #NOTPATCH
        if newKeywords is None:
            LOG('newKeywords',100,newKeywords)
            return 0
        #NOTPATCH
        if callable(newKeywords):
            newKeywords = newKeywords()

        if type(newKeywords) is StringType:
            newKeywords = (newKeywords, )

        oldKeywords = self._unindex.get(documentId, None)
        
        if oldKeywords is None:
            # we've got a new document, let's not futz around.
            try:
                for kw in newKeywords:
                    self.insertForwardIndexEntry(kw, documentId)
                self._unindex[documentId] = list(newKeywords)
            except TypeError:
                return 0
        else:
            # we have an existing entry for this document, and we need
            # to figure out if any of the keywords have actually changed
            if type(oldKeywords) is not OOSet: oldKeywords=OOSet(oldKeywords)
            newKeywords=OOSet(newKeywords)
            fdiff = difference(oldKeywords, newKeywords)
            rdiff = difference(newKeywords, oldKeywords)
            if fdiff or rdiff:
                # if we've got forward or reverse changes
                self._unindex[documentId] = list(newKeywords)
                if fdiff:
                    self.unindex_objectKeywords(documentId, fdiff)
                if rdiff:
                    for kw in rdiff:
                        self.insertForwardIndexEntry(kw, documentId)
        
        return 1

    def unindex_objectKeywords(self, documentId, keywords):
        """ carefully unindex the object with integer id 'documentId'"""

        if keywords is not None:
            for kw in keywords:
                self.removeForwardIndexEntry(kw, documentId)

    def unindex_object(self, documentId):
        """ carefully unindex the object with integer id 'documentId'"""

        keywords = self._unindex.get(documentId, None)
        self.unindex_objectKeywords(documentId, keywords)
        try:
            del self._unindex[documentId]
        except KeyError:
            LOG('UnKeywordIndex', ERROR, 'Attempt to unindex nonexistent'
                ' document id %s' % documentId)


    
    manage_workspace = DTMLFile('dtml/manageEasyKeywordIndex', globals())


    def _apply_index(self, request, cid='', type=type, None=None): 
        """Apply the index to query parameters given in the request arg.

        The request argument should be a mapping object.

        If the request does not have a key which matches the "id" of
        the index instance, then None is returned.

        If the request *does* have a key which matches the "id" of
        the index instance, one of a few things can happen:

          - if the value is a blank string, None is returned (in
            order to support requests from web forms where
            you can't tell a blank string from empty).

          - if the value is a nonblank string, turn the value into
            a single-element sequence, and proceed.

          - if the value is a sequence, return a union search.

        If the request contains a parameter with the name of the
        column + '_usage', it is sniffed for information on how to
        handle applying the index.

        If the request contains a parameter with the name of the 
        column = '_operator' this overrides the default method
        ('or') to combine search results. Valid values are "or"
        and "and".

        If None is not returned as a result of the abovementioned
        constraints, two objects are returned.  The first object is a
        ResultSet containing the record numbers of the matching
        records.  The second object is a tuple containing the names of
        all data fields used.

        FAQ answer:  to search a Field Index for documents that
        have a blank string as their value, wrap the request value
        up in a tuple ala: request = {'id':('',)}

        """
        

        record = parseIndexRequest(request, self.id, self.query_options)
        if record.keys==None: return None

        index = self._index
        r     = None
        opr   = None


        # experimental code for specifing the operator
        operator = record.get('operator',self.useOperator)
        if not operator in self.operators :
           raise RuntimeError,"operator not valid: %s" % operator

        # depending on the operator we use intersection or union
        if operator=="or":          set_func = IIunion
        elif operator=="not":       set_func = IIunion          #NOTPATCH
        elif operator=="notor":     set_func = IIunion          #NOTPATCH
        elif operator=="notand":    set_func = IIintersection   #NOTPATCH
        else:                       set_func = IIintersection

        # Range parameter
        range_parm = record.get('range',None)
        if range_parm:
            opr = "range"
            opr_args = []
            if range_parm.find("min")>-1:
                opr_args.append("min")
            if range_parm.find("max")>-1:
                opr_args.append("max")


        if record.get('usage',None):
            # see if any usage params are sent to field
            opr = record.usage.lower().split(':')
            opr, opr_args=opr[0], opr[1:]

        
        if opr=="range":   # range search
            if 'min' in opr_args: lo = min(record.keys)
            else: lo = None
            if 'max' in opr_args: hi = max(record.keys)
            else: hi = None
            if hi:
                setlist = index.items(lo,hi)
            else:
                setlist = index.items(lo)

            for k, set in setlist:
                if type(set) is IntType:
                    set = IISet((set,))
                r = set_func(r, set)
        else: # not a range search
            for key in record.keys:
                set=index.get(key, None)
                if set is not None:
                    if type(set) is IntType:
                        set = IISet((set,))
                    r = set_func(r, set)

        if type(r) is IntType:  r=IISet((r,))
        
        #NOTPATCH
        if operator=="not" or operator=="notor" or operator=="notand": 
            r = IIdifference( IISet(self._unindex.keys()), r) 
        #NOTPATCH
        
        if r is None:
            return IISet(), (self.id,)
        else:
            return r, (self.id,)


manage_addEasyKeywordIndexForm = DTMLFile('dtml/addEasyKeywordIndex', globals())

def manage_addEasyKeywordIndex(self, id, REQUEST=None, RESPONSE=None, URL3=None):
    """Add a keyword index"""
    return self.manage_addIndex(id, 'EasyKeywordIndex', extra=None, \
              REQUEST=REQUEST, RESPONSE=RESPONSE, URL1=URL3)




